import pytest

from yatbaf.types import InlineKeyboardButton
from yatbaf_menu import Back
from yatbaf_menu import Menu
from yatbaf_menu import Submenu


def test_init_no_parent(router, payload):
    menu = Menu(
        title="menu",
        name="menu",
        buttons=[
            button := Back(title="Back"),
        ],
    )
    with pytest.raises(ValueError):
        button._init(menu, router, payload)


def test_init(router, prefix, payload):
    menu = Menu(
        title="menu",
        name="menu",
        submenu=[
            submenu := Menu(
                title="submenu",
                name="submenu",
                buttons=[
                    button := Back(title="back"),
                ],
            ),
        ],
        buttons=[Submenu("button", "submenu")],
    )
    menu._prefix = prefix
    submenu._parent = menu
    button._init(submenu, router, payload)
    assert button._payload == f"{menu._prefix}##"
    assert button._show is None


def test_init_duplicate(router, payload):
    menu = Menu(
        title="menu",
        name="menu",
        submenu=[
            submenu := Menu(
                title="submenu",
                name="submenu",
                buttons=[
                    button := Back(title="back"),
                ],
            ),
        ],
        buttons=[Submenu("button", "submenu")],
    )
    button._payload = "qwe"
    submenu._parent = menu
    with pytest.raises(ValueError):
        button._init(submenu, router, payload)


@pytest.mark.asyncio
async def test_build(callback_query, prefix):
    button = Back(title="back")
    button._payload = f"{prefix}##"
    result = await button._build(callback_query)
    assert result == InlineKeyboardButton(
        text="back",
        callback_data=button._payload,
    )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import TypeAlias

if TYPE_CHECKING:
    from collections.abc import Awaitable
    from collections.abc import Callable
    from collections.abc import Iterable

    from yatbaf.types import CallbackQuery
    from yatbaf.types import Message

Query: TypeAlias = "Message | CallbackQuery"
ChoiceItemsFn: TypeAlias = (
    "Callable[[Query], Awaitable[Iterable[tuple[str, str]]]]"
)
ChoiceItems: TypeAlias = "list[tuple[str, str]]"
PagingItemsFn: TypeAlias = (
    "Callable[[Query, int, int], Awaitable[tuple[Iterable[tuple[str, str]], bool]]]"  # noqa: E501
)



__all__ = (
    "URL",
    "Action",
    "Back",
    "Submenu",
    "Menu",
    "build_router",
    "Choice",
    "Paging",
)

from .button import URL
from .button import Action
from .button import Back
from .button import Submenu
from .menu import Menu
from .row import Choice
from .row import Paging
from .utils import build_router



import pytest

from yatbaf.router import OnCallbackQuery
from yatbaf.types import CallbackQuery
from yatbaf.types import User
from yatbaf_menu.payload import Payload


@pytest.fixture
def router():
    return OnCallbackQuery()


@pytest.fixture
def prefix():
    return "aa00aa"


@pytest.fixture
def payload(prefix):
    return Payload(prefix)


@pytest.fixture
def user():
    return User(
        id=12345,
        is_bot=False,
        first_name="Test",
    )


@pytest.fixture
def callback_query(user):
    return CallbackQuery(
        id=12345,
        chat_instance="test",
        from_=user,
    )



import pytest

from yatbaf.types import InlineKeyboardButton
from yatbaf_menu import Action


def test_init(router, menu, payload):
    action = object()
    button = Action(title="button", action=action)
    button._init(menu, router, payload)
    assert button._payload is not None
    assert router._handlers[0]._fn is action


def test_init_duplicate(router, menu, payload):
    action = object()
    button = Action(title="button", action=action)
    button._payload = "aa00aa"
    with pytest.raises(ValueError):
        button._init(menu, router, payload)


def test_init_duplicate_action(router, menu, payload):
    action = object()
    button1 = Action(title="button1", action=action)
    button1._init(menu, router, payload)
    button2 = Action(title="button2", action=action)
    with pytest.raises(ValueError):
        button2._init(menu, router, payload)


@pytest.mark.asyncio
async def test_build(callback_query):
    paylaod = "aa00aa"
    button = Action(title="button", action=object())
    button._payload = paylaod

    result = await button._build(callback_query)
    assert result == InlineKeyboardButton(text="button", callback_data=paylaod)


@pytest.mark.asyncio
async def test_build_not_visible(callback_query):

    async def _visible(_):
        return False

    paylaod = "aa00aa"
    button = Action(title="button", action=object(), show=_visible)
    button._payload = paylaod

    result = await button._build(callback_query)
    assert result is None


@pytest.mark.asyncio
async def test_build_extra_payload(callback_query):
    paylaod = "aa00aa"
    extra = "some-payload"

    def _payload(_):
        return extra

    button = Action(title="button", action=object(), payload=_payload)
    button._payload = paylaod

    result = await button._build(callback_query)
    assert result == InlineKeyboardButton(
        text="button",
        callback_data=f"{paylaod}{extra}",
    )



from __future__ import annotations

__all__ = (
    "AbstractButton",
    "Action",
    "URL",
    "Submenu",
    "Back",
)

import asyncio
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

from yatbaf.handler import Handler
from yatbaf.types import InlineKeyboardButton

from .filter import CallbackPayload

if TYPE_CHECKING:
    from collections.abc import Awaitable
    from collections.abc import Callable

    from yatbaf import OnCallbackQuery

    from .menu import Menu
    from .payload import Payload
    from .typing import Query


class AbstractButton(ABC):
    __slots__ = ()

    @abstractmethod
    def _init(
        self, menu: Menu, router: OnCallbackQuery, payload: Payload, /
    ) -> None:
        pass

    @abstractmethod
    async def _build(self, q: Query, /) -> InlineKeyboardButton | None:
        pass


class BaseButton(AbstractButton):
    __slots__ = (
        "_title",
        "_show",
    )

    def __init__(
        self,
        *,
        title: str | Callable[[Query], Awaitable[str]],
        show: Callable[[Query], Awaitable[bool]] | None = None,
    ) -> None:
        self._title = title
        self._show = show

    def __repr__(self) -> str:
        title = "`dynamic`" if callable(t := self._title) else t
        return f"<{self.__class__.__name__}[{title=!r}]>"

    async def _get_title(self, q: Query, /) -> str:
        if callable(self._title):
            return await self._title(q)
        return self._title

    async def _is_visible(self, q: Query, /) -> bool:
        if self._show is not None:
            return await self._show(q)
        return True


class CallbackButton(BaseButton):
    __slots__ = (
        "_payload",
        "_payload_extra",
    )

    def __init__(
        self,
        title: str | Callable[[Query], Awaitable[str]],
        show: Callable[[Query], Awaitable[bool]] | None = None,
        payload: Callable[[Query], str] | None = None,
    ) -> None:
        super().__init__(
            title=title,
            show=show,
        )
        self._payload: str | None = None
        self._payload_extra = payload

    def __eq__(self, other: object) -> bool:
        return other is self or (  # yapf: disable
            isinstance(other, self.__class__) and (
                other._title == self._title
                and other._show == self._show
                and other._payload_extra == self._payload_extra
            )
        )

    async def _build(self, q: Query) -> InlineKeyboardButton | None:
        if not await self._is_visible(q):
            return None

        extra = "" if (f := self._payload_extra) is None else f(q)
        return InlineKeyboardButton(
            text=await self._get_title(q),
            callback_data=f"{self._payload}{extra}",
        )


class Action(CallbackButton):
    """This button does the action"""
    __slots__ = ("_action",)

    def __init__(
        self,
        title: str | Callable[[Query], Awaitable[str]],
        action: Callable[[Query], Awaitable[None]],
        *,
        show: Callable[[Query], Awaitable[bool]] | None = None,
        payload: Callable[[Query], str] | None = None,
    ) -> None:
        """
        :param title: String or Callable which returns button title.
        :param action: Callable to run on click. Must be unique for the menu.
        :param show: *Optional.* Callable which returns visibility status.
        :param payload: *Optional.* Sync function, returned string will be used
            as `callback_data`.
        """
        super().__init__(
            title=title,
            show=show,
            payload=payload,
        )
        self._action = action

    def __eq__(self, other: object) -> bool:
        return (
            super().__eq__(other)
            and other._action == self._action  # type: ignore[attr-defined]
        )

    def _init(
        self, m: Menu, router: OnCallbackQuery, payload: Payload, /
    ) -> None:
        if self._payload is not None:
            raise ValueError(
                f"{self!r} button must be unique to the entire menu."
            )

        self._payload = payload.get()
        handler = Handler(
            self._action,
            update_type=router._update_type,
            filters=[CallbackPayload(self._payload)],
        )
        if handler in router._handlers:
            raise ValueError(f"{self!r} `action` must be unique to menu.")

        router.add_handler(handler)


class URL(BaseButton):
    """This button will open URL"""
    __slots__ = ("_url",)

    def __init__(
        self,
        title: str | Callable[[Query], Awaitable[str]],
        url: str | Callable[[Query], Awaitable[str]],
        *,
        show: Callable[[Query], Awaitable[bool]] | None = None,
    ) -> None:
        """
        :param url: String or Callable which returns url.
        :param title: String or Callable which returns button title.
        :param show: *Optional.* Callable which returns visibility status.
        """
        super().__init__(
            title=title,
            show=show,
        )
        self._url = url

    def __eq__(self, other: object) -> bool:
        return isinstance(other, URL) and (  # yapf: disable
            other is self or (
                other._title == self._title
                and other._url == self._url
                and other._show == self._show
            )
        )

    def _init(self, m: Menu, r: OnCallbackQuery, p: Payload, /) -> None:
        pass  # nothing to do

    async def _get_url(self, q: Query, /) -> str:
        if callable(self._url):
            return await self._url(q)
        return self._url

    async def _build(self, q: Query, /) -> InlineKeyboardButton | None:
        if not await self._is_visible(q):
            return None

        async with asyncio.TaskGroup() as tg:
            title = tg.create_task(self._get_title(q))
            url = tg.create_task(self._get_url(q))

        return InlineKeyboardButton(
            text=title.result(),
            url=url.result(),
        )


class Submenu(CallbackButton):
    """This button will open next menu"""
    __slots__ = ("_menu",)

    def __init__(
        self,
        title: str | Callable[[Query], Awaitable[str]],
        menu: str,
        *,
        show: Callable[[Query], Awaitable[bool]] | None = None,
        payload: Callable[[Query], str] | None = None,
    ) -> None:
        """
        :param menu: Submenu name (see :class:`~yatbaf_menu.menu.Menu`).
        :param title: String or Callable which returns button title.
        :param show: *Optional.* Callable which returns visibility status.
        :param payload: *Optional.* Sync function, returned string will be used
            as `callback_data`.
        """
        super().__init__(
            title=title,
            show=show,
            payload=payload,
        )
        self._menu = menu

    def __repr__(self) -> str:
        title = "`dynamic`" if callable(t := self._title) else t
        menu = self._menu
        return f"<Submenu[{title=}, {menu=}]>"

    def __eq__(self, other: object) -> bool:
        return (
            super().__eq__(other)
            and other._menu == self._menu  # type: ignore[attr-defined]
        )

    def _init(self, menu: Menu, r: OnCallbackQuery, p: Payload, /) -> None:
        if self._payload is not None:
            raise ValueError(f"{self!r} button must be unique.")
        self._payload = f"{menu.get_submenu(self._menu)._prefix}##"


class Back(CallbackButton):
    """This button will open previous menu"""
    __slots__ = ()

    def __init__(
        self,
        title: str | Callable[[Query], Awaitable[str]],
        *,
        payload: Callable[[Query], str] | None = None,
    ) -> None:
        """
        :param title: String or Callable which returns button title.
        :param payload: *Optional.* Sync function, returned string will be used
            as `callback_data`.
        """
        super().__init__(
            title=title,
            payload=payload,
        )

    def _init(self, menu: Menu, r: OnCallbackQuery, p: Payload, /) -> None:
        if self._payload is not None:
            raise ValueError(f"{self!r} button must be unique.")

        if menu._parent is None:
            raise ValueError(f"It is not possible to use {self!r} in {menu!r}.")

        self._payload = f"{menu._parent._prefix}##"



import pytest

from yatbaf_menu.menu import Menu


@pytest.fixture
def menu(prefix):
    menu_ = Menu(
        title="menu",
        name="menu",
        back_btn_title="back",
    )
    menu_._prefix = prefix
    return menu_



from yatbaf_menu.button import Action
from yatbaf_menu.button import Back
from yatbaf_menu.button import Submenu
from yatbaf_menu.menu import Menu
from yatbaf_menu.row import Row
from yatbaf_menu.utils import build_router


async def action1(_):
    pass


async def action2(_):
    pass


def test_build():
    menu = Menu(
        title="title",
        name="main",
        buttons=[
            Submenu(title="sub1", menu="sub1"),
            Action(title="button2", action=action2),
        ],
        submenu=[
            Menu(
                title="title",
                name="sub1",
                buttons=[Submenu(title="sub2", menu="sub2")],
                submenu=[
                    Menu(
                        title="title",
                        name="sub2",
                        back_btn_title="back",
                    ),
                ],
                back_btn_title="back",
            ),
            Menu(
                title="title",
                name="sub3",
                buttons=[Action(title="button1", action=action1)],
                back_btn_title="back",
            )
        ],
    )

    router = build_router(menu)
    """
    root-router
      ├── base-router
      │     └── menu-main
      │           ├── menu-sub1
      │           │     └── menu-sub2
      │           └── menu-sub3
      └── fallback-router
    """

    # yapf: disable
    # root router
    assert len(router._routers) == 2  # base, fallback
    assert not router._handlers

    # base router
    assert len(router._routers[0]._routers) == 1  # main

    # main menu
    assert len(router._routers[0]._routers[0]._routers) == 2  # sub1, sub3
    assert len(router._routers[0]._routers[0]._handlers) == 2  # submenu, action
    assert router._routers[0]._routers[0]._handlers[1]._fn is action2

    # sub1 router
    assert len(router._routers[0]._routers[0]._routers[0]._routers) == 1  # sub2
    assert len(router._routers[0]._routers[0]._routers[0]._handlers) == 1  # back  # noqa: E501

    # sub2 router
    assert not router._routers[0]._routers[0]._routers[0]._routers[0]._routers
    assert len(router._routers[0]._routers[0]._routers[0]._routers[0]._handlers) == 1  # back  # noqa: E501

    # sub3 router
    assert not router._routers[0]._routers[0]._routers[1]._routers
    assert len(router._routers[0]._routers[0]._routers[1]._handlers) == 2  # action, back  # noqa: E501
    assert router._routers[0]._routers[0]._routers[1]._handlers[1]._fn is action1  # noqa: E501

    # fallback router
    assert not router._routers[1]._routers
    assert len(router._routers[1]._handlers) == 1
    # yapf: enable

    assert menu._parent is None
    assert menu._prefix is not None
    assert menu._rows == [
        Row([Submenu(title="sub1", menu="sub1")]),
        Row([Action(title="button2", action=action2)]),
    ]

    sub1 = menu._submenu["sub1"]
    assert sub1._parent is menu
    assert sub1._prefix is not None
    assert sub1._rows == [
        Row([Submenu(title="sub2", menu="sub2")]),
        Row([Back(title="back")]),
    ]

    sub2 = sub1._submenu["sub2"]
    assert sub2._parent is sub1
    assert sub2._prefix is not None
    assert sub2._rows == [
        Row([Back(title="back")]),
    ]

    sub3 = menu._submenu["sub3"]
    assert sub3._parent is menu
    assert sub3._prefix is not None
    assert sub3._rows == [
        Row([Action(title="button1", action=action1)]),
        Row([Back(title="back")]),
    ]



import pytest

from yatbaf.types import InlineKeyboardButton
from yatbaf_menu.menu import Menu
from yatbaf_menu.row import Paging

PAYLOAD = "~a00a"
NAV_PAYLOAD = "~a00a"
NEXT_BTN_TITLE = "next"
PREV_BTN_TITLE = "next"


def test_no_action_and_submenu():
    with pytest.raises(ValueError):
        Paging(items=object())


def test_action_and_submenu():
    with pytest.raises(ValueError):
        Paging(action=object(), submenu="menu", items=object())


def test_init_submenu(prefix, router, payload):
    menu = Menu(
        title="menu",
        name="menu",
        submenu=[
            submenu := Menu(
                title="submenu",
                name="submenu",
                back_btn_title="back",
            ),
        ],
        buttons=[
            paging := Paging(
                submenu="submenu",
                items=object(),
            ),
        ],
    )
    menu._prefix = prefix
    submenu._prefix = f"{prefix[:-2]}bb"
    paging._init(menu, router, payload)
    assert paging._payload == f"{submenu._prefix}##"
    assert paging._nav_payload == f"{menu._prefix}##"
    assert len(router._handlers) == 0


def test_init_action(router, payload, prefix):
    action = object()
    menu = Menu(
        title="menu",
        name="menu",
        buttons=[
            paging := Paging(action=action, items=object()),
        ],
    )
    menu._prefix = prefix
    paging._init(menu, router, payload)
    assert paging._payload is not None
    assert paging._nav_payload == f"{menu._prefix}##"
    assert len(router._handlers) == 1
    assert router._handlers[0]._fn is action


def test_init_duplicate(router, payload, prefix):
    menu = Menu(
        title="menu",
        name="menu",
        buttons=[
            paging := Paging(action=object(), items=object()),
        ],
    )
    menu._prefix = prefix
    paging._payload = "qwe"
    with pytest.raises(ValueError):
        paging._init(menu, router, payload)


def test_init_duplicate_action(router, payload, prefix):
    action = object()
    menu = Menu(
        title="menu",
        name="menu",
        buttons=[
            paging := Paging(action=action, items=object()),
        ],
    )
    menu._prefix = prefix
    router.add_handler(action)
    with pytest.raises(ValueError):
        paging._init(menu, router, payload)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "offset,last_page,expected",
    [  # yapf: disable
        (
            0, False,
            [
                [InlineKeyboardButton(text="1", callback_data=f"{PAYLOAD}1")],
                [InlineKeyboardButton(text="2", callback_data=f"{PAYLOAD}2")],
                [InlineKeyboardButton(text=f"{NEXT_BTN_TITLE}", callback_data=f"{NAV_PAYLOAD}#2")],  # noqa: E501
            ],
        ),
        (
            2, False,
            [
                [InlineKeyboardButton(text="1", callback_data=f"{PAYLOAD}1")],
                [InlineKeyboardButton(text="2", callback_data=f"{PAYLOAD}2")],
                [
                    InlineKeyboardButton(text=f"{PREV_BTN_TITLE}", callback_data=f"{NAV_PAYLOAD}#0"),  # noqa: E501
                    InlineKeyboardButton(text=f"{NEXT_BTN_TITLE}", callback_data=f"{NAV_PAYLOAD}#4"),  # noqa: E501
                ],
            ],
        ),
        (
            2, True,
            [
                [InlineKeyboardButton(text="1", callback_data=f"{PAYLOAD}1")],
                [InlineKeyboardButton(text="2", callback_data=f"{PAYLOAD}2")],
                [InlineKeyboardButton(text=f"{PREV_BTN_TITLE}", callback_data=f"{NAV_PAYLOAD}#0")],  # noqa: E501
            ],
        ),
    ]
)
async def test_build(callback_query, offset, last_page, expected):

    async def _items(q, o, i):  # noqa: U100
        return ([("1", "1"), ("2", "2")], last_page)

    paging = Paging(
        action=object(),
        items=_items,
        items_per_row=1,
        items_per_page=2,
        prev_btn_title=PREV_BTN_TITLE,
        next_btn_title=NEXT_BTN_TITLE,
    )
    paging._payload = PAYLOAD
    paging._nav_payload = NAV_PAYLOAD
    callback_query.data = f"#{offset}"
    result = await paging._build(callback_query)
    assert result == expected


@pytest.mark.parametrize(
    "payload,expected",
    [
        ("12", 0),
        ("", 0),
        ("#12", 12),
        ("#-12", 0),
        ("payload#12", 12),
        ("data#payload#12", 12),
    ]
)
def test_parse_offset(payload, expected):
    assert Paging._parse_offset(payload) == expected



import pytest

from yatbaf.types import InlineKeyboardButton
from yatbaf_menu.button import URL
from yatbaf_menu.row import Row


@pytest.mark.asyncio
async def test_row():
    row = Row([
        URL(url="example.com", title="example 1"),
        URL(url="example.com", title="example 2"),
    ])

    result = await row._build(None)
    assert result == [
        [
            InlineKeyboardButton(url="example.com", text="example 1"),
            InlineKeyboardButton(url="example.com", text="example 2")
        ],
    ]



from __future__ import annotations

__all__ = ("build_router",)

from typing import TYPE_CHECKING
from typing import cast

from yatbaf import OnCallbackQuery
from yatbaf.exceptions import SkipRouterException
from yatbaf.middleware import Middleware

from .filter import CallbackPayload
from .middleware import CutPayloadMiddileware
from .middleware import InjectMenuMiddleware
from .payload import Payload

if TYPE_CHECKING:
    from yatbaf.types import CallbackQuery

    from .menu import Menu
    from .nav import MenuNav

_root_payload_bucket = Payload()


def _build_menu_router(menu: Menu) -> OnCallbackQuery:
    """
    :meta private:

    :param menu: Menu object.
    """
    menu._prefix = cast("str", menu._prefix)
    payload = menu._prefix[3:5]

    def _guard(q: CallbackQuery) -> None:
        if q.data[3:5] == payload:  # type: ignore[index]
            return
        raise SkipRouterException

    menu._guards.insert(0, _guard)
    router = OnCallbackQuery(
        name=menu.name,
        guards=menu._guards,
        middleware=menu._middleware,
    )

    @router(filters=[CallbackPayload(f"{menu._prefix}##")])
    async def _open_menu(q: CallbackQuery) -> None:
        menu: MenuNav = q.ctx["menu"]
        await q.answer()
        await menu.refresh()

    return router


def _build_base_router(menu: Menu, version: str) -> OnCallbackQuery:
    """
    :meta private:

    :param menu: Main menu object.
    """
    router = OnCallbackQuery(
        name=f"{menu.name}-root",
        middleware=[CutPayloadMiddileware],
    )

    @router.guard
    def _guard(q: CallbackQuery) -> None:
        if q.data[2:4] == version:  # type: ignore[index]
            return
        raise SkipRouterException(skip_nested=True)

    return router


def _build_fallback_router(menu: Menu, update_message: str) -> OnCallbackQuery:
    """
    :meta private:

    :param menu: Main menu object.
    :param update_message: Will be shown to the user.
    """
    router = OnCallbackQuery(
        name=f"{menu.name}-fallback",
        middleware=[
            Middleware(
                InjectMenuMiddleware,  # type: ignore[arg-type]
                is_local=True,
                is_handler=True,
                menu=menu,
            ),
        ],
    )

    @router
    async def _fallback(q: CallbackQuery) -> None:
        menu: MenuNav = q.ctx["menu"]
        q.data = None
        await q.answer(text=update_message)
        await menu.open_main()

    return router


def _build_root_router(
    menu: Menu,
    base_router: OnCallbackQuery,
    fallback_router: OnCallbackQuery,
) -> OnCallbackQuery:
    """:meta private:"""
    menu._prefix = cast("str", menu._prefix)
    prefix = menu._prefix[:2]  # root prefix 2 chars

    router = OnCallbackQuery(
        name=f"{menu.name}-base",
        routers=[
            base_router,
            fallback_router,
        ],
    )

    @router.guard
    def _guard(q: CallbackQuery) -> None:
        if (data := q.data) is not None and data[:2] == prefix:
            return
        raise SkipRouterException(skip_nested=True)

    return router


def _init_buttons(menu: Menu, router: OnCallbackQuery) -> None:
    menu._prefix = cast("str", menu._prefix)
    payload = Payload(menu._prefix)
    for row in menu._rows:
        row._init(menu, router, payload)


def _parse_version(version: str | int) -> str:
    version = str(version)
    if len(version) > 2:
        raise ValueError(f"{version=} cannot be longer than 2 characters.")
    return version.zfill(2)


def build_router(
    menu: Menu,
    *,
    version: str | int = 0,
    update_message: str = "This menu was updated.",
) -> OnCallbackQuery:
    """Build menu router.

    :param menu: :class:`~yatbaf_menu.menu.Menu` object.
    :param version: *Optional.* Menu version. Up to 2 chars.
    :param update_message: *Optional.* Will be shown to the user when
        ``version`` is increased.
    :returns: Configured :class:`~yatbaf.router.OnCallbackQuery` object.
    """
    version = _parse_version(version)
    menu_payload = Payload(_root_payload_bucket.get() + version)
    base_router = _build_base_router(menu, version)

    def _init_menu(menu: Menu, base: OnCallbackQuery) -> None:
        menu._prefix = menu_payload.get()
        router = _build_menu_router(menu)
        base.add_router(router)
        for submenu in menu._submenu.values():
            submenu._parent = menu
            _init_menu(submenu, router)
        _init_buttons(menu, router)

    _init_menu(menu, base_router)

    return _build_root_router(
        menu=menu,
        base_router=base_router,
        fallback_router=_build_fallback_router(
            menu=menu,
            update_message=update_message,
        ),
    )



from __future__ import annotations

__all__ = ("CallbackPayload",)

from typing import TYPE_CHECKING
from typing import final

if TYPE_CHECKING:
    from yatbaf.types import CallbackQuery


@final
class CallbackPayload:
    __slots__ = (
        "payload",
        "priority",
    )

    def __init__(self, payload: str) -> None:
        self.payload = payload
        self.priority = 300

    def check(self, q: CallbackQuery) -> bool:
        return q.data[:8] == self.payload  # type: ignore[index]



from __future__ import annotations

__all__ = ("Menu",)

import asyncio
from itertools import count
from typing import TYPE_CHECKING
from typing import Any
from typing import cast
from typing import final

from yatbaf.middleware import Middleware
from yatbaf.types import InlineKeyboardMarkup

from .button import Back
from .middleware import InjectMenuMiddleware
from .row import AbstractRowBuilder
from .row import Row

if TYPE_CHECKING:
    from collections.abc import Awaitable
    from collections.abc import Callable
    from collections.abc import Sequence

    from yatbaf.enums import ParseMode
    from yatbaf.router import InitMiddlewareType
    from yatbaf.types import CallbackQuery
    from yatbaf.typing import GuardCallable

    from .button import AbstractButton
    from .typing import Query

_menu_count = count(1).__next__


def _parse_markup(
    markup: list[AbstractButton | AbstractRowBuilder | list[AbstractButton]]
) -> list[AbstractRowBuilder]:
    result: list[AbstractRowBuilder] = []
    for row in markup:
        if isinstance(row, list):
            for obj in row:
                if issubclass(type(obj), AbstractRowBuilder):
                    if len(row) != 1:
                        raise ValueError(
                            f"It is not possible to use {obj!r} with "
                            "other objects in the same row"
                        )
                    result.append(cast("AbstractRowBuilder", obj))
                    break
            else:
                result.append(Row(row))

        # row builder
        elif issubclass(type(row), AbstractRowBuilder):
            result.append(cast("AbstractRowBuilder", row))

        # single button
        else:
            result.append(Row([cast("AbstractButton", row)]))

    return result


def _parse_menu_name(name: str | None) -> str:
    if name is not None:
        return name.lower()
    return f"menu{_menu_count()}"


@final
class Menu:
    """Menu object."""

    __slots__ = (
        "_title",
        "_name",
        "_buttons",
        "_submenu",
        "_rows",
        "_parent",
        "_prefix",
        "_guards",
        "_middleware",
        "_parse_mode",
    )

    def __init__(  # yapf: disable
        self,
        *,
        title: str | Callable[[Query], Awaitable[str]],
        name: str | None = None,
        buttons: list[AbstractButton | AbstractRowBuilder | list[AbstractButton]] | None = None,  # noqa: E501
        submenu: Sequence[Menu] | None = None,
        guards: Sequence[GuardCallable[CallbackQuery]] | None = None,
        middleware: Sequence[InitMiddlewareType[CallbackQuery]] | None = None,
        back_btn_title: str | Callable | None = None,
        parse_mode: ParseMode | None = None,
    ) -> None:
        """
        :param title: String or Callable which return menu title.
        :param name: *Optional.* Menu name.
        :param submenu: *Optional.* Sequence of :class:`Menu`.
        :param buttons: *Optional.* A list of buttons.
        :param parse_mode: *Optional.* Parse mode for menu title.
        :param back_btn_title: *Optional.* Pass a title if you want to add
            a 'back button' to this menu. For submenu only.
        :param guards: *Optional.* Sequence of :class:`~yatbaf.typing.Guard`
        :param middleware: *Optional.* Sequence of :class:`~yatbaf.typing.Middleware`
        """  # noqa: E501
        self._name = _parse_menu_name(name)
        self._title = title
        self._prefix: str | None = None

        buttons = buttons if buttons is not None else []
        if back_btn_title is not None:
            buttons.append(Back(title=back_btn_title))
        if not buttons:
            raise ValueError(f"{self!r} must have at least one button.")

        self._rows = _parse_markup(buttons)

        self._submenu = {} if submenu is None else {m.name: m for m in submenu}
        self._parent: Menu | None = None

        self._guards = list(guards) if guards is not None else []
        self._middleware = list(middleware) if middleware is not None else []
        self._middleware.append(
            Middleware(
                InjectMenuMiddleware,  # type: ignore[arg-type]
                is_local=True,
                is_handler=True,
                menu=self,
            )
        )
        self._parse_mode = parse_mode

    def __repr__(self) -> str:
        return f"<Menu[{self._name}]>"

    @property
    def name(self) -> str:
        """Menu name."""
        return self._name

    def get_submenu(self, name: str) -> Menu:
        """Returns submenu with by name.

        :param name: Submenu name.
        :raises ValueError: If submenu with ``name`` is not found.
        """
        try:
            return self._submenu[name]
        except KeyError:
            raise ValueError(f"Menu {name} not found in {self!r}") from None

    def get_root(self) -> Menu:
        """Returns main menu."""
        menu = self
        while menu._parent is not None:
            menu = menu._parent
        return menu

    def get_menu(self, path: str) -> Menu:
        """Returns menu by path.

        :param path: Path to menu.
        :raises ValueError: If ``path`` is not found.
        """
        menu = self.get_root()
        for name in path.split("."):
            menu = menu.get_submenu(name)
        return menu

    async def _get_title(self, q: Query, /) -> str:
        if callable(self._title):
            return await self._title(q)
        return self._title

    async def _get_markup(self, q: Query, /) -> InlineKeyboardMarkup:
        async with asyncio.TaskGroup() as tg:
            tasks = [tg.create_task(r._build(q)) for r in self._rows]
        return InlineKeyboardMarkup([
            row for task in tasks for row in task.result()
        ])

    async def get_message_params(self, query: Query) -> dict[str, Any]:
        """Create parameters for message.

        Use it to open menu::

            @on_message(filters=[Command("menu")])
            async def open_menu(message: Message) -> None:
                params = await menu.get_message_params(message)
                await message.answer(**params)

        :param query: :class:`~yatbaf.types.message.Message` or
            :class:`~yatbaf.types.callback_query.CallbackQuery` object.
        """
        async with asyncio.TaskGroup() as tg:
            title = tg.create_task(self._get_title(query))
            markup = tg.create_task(self._get_markup(query))

        return {
            "text": title.result(),
            "reply_markup": markup.result(),
            "parse_mode": self._parse_mode,
        }



from __future__ import annotations

__all__ = (
    "AbstractRowBuilder",
    "Choice",
    "Paging",
    "Row",
)

import asyncio
from abc import ABC
from abc import abstractmethod
from itertools import islice
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import cast

from yatbaf.handler import Handler
from yatbaf.types import InlineKeyboardButton

from .filter import CallbackPayload

if TYPE_CHECKING:
    from collections.abc import Iterable
    from collections.abc import Iterator

    from yatbaf.router import OnCallbackQuery
    from yatbaf.types import CallbackQuery
    from yatbaf.typing import HandlerCallable

    from .button import AbstractButton
    from .menu import Menu
    from .payload import Payload
    from .typing import ChoiceItems
    from .typing import ChoiceItemsFn
    from .typing import PagingItemsFn
    from .typing import Query


class AbstractRowBuilder(ABC):
    __slots__ = ()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    @abstractmethod
    def _init(
        self,
        menu: Menu,
        router: OnCallbackQuery,
        payload: Payload,
    ) -> None:
        pass

    @abstractmethod
    async def _build(self, q: Query) -> list[list[InlineKeyboardButton]]:
        pass


class Row(AbstractRowBuilder):
    __slots__ = ("_buttons",)

    def __init__(self, buttons: list[AbstractButton]) -> None:
        self._buttons = buttons

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Row)
            and (other is self or (other._buttons == self._buttons))
        )

    def _init(
        self,
        menu: Menu,
        router: OnCallbackQuery,
        payload: Payload,
    ) -> None:
        for button in self._buttons:
            button._init(menu, router, payload)

    async def _build(self, q: Query) -> list[list[InlineKeyboardButton]]:
        async with asyncio.TaskGroup() as tg:
            tasks = [tg.create_task(b._build(q)) for b in self._buttons]
        if row := [b for t in tasks if (b := t.result()) is not None]:
            return [row]
        return row  # type: ignore[return-value]


T = TypeVar("T")


def batched(iterable: Iterable[T], n: int) -> Iterator[tuple[T, ...]]:
    """:meta private:"""
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError('n must be at least one')
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


def build_rows(
    items: Iterable[tuple[str, str]],
    payload: str,
    per_row: int,
) -> list[list[InlineKeyboardButton]]:
    """:meta private:"""
    return [  # yapf: disable
        [
            InlineKeyboardButton(
                text=title,
                callback_data=f"{payload}{extra}",
            )
            for title, extra in row
        ]
        for row in batched(items, per_row)
    ]


class Choice(AbstractRowBuilder):
    """Choice row builder."""

    __slots__ = (
        "_action",
        "_items",
        "_items_per_row",
        "_payload",
        "_submenu",
    )

    def __init__(
        self,
        items: ChoiceItems | ChoiceItemsFn | Iterable[str | int],
        action: HandlerCallable[CallbackQuery] | None = None,
        submenu: str | None = None,
        items_per_row: int = 1,
    ) -> None:
        """
        :param items: Callable or list of items.
        :param action: *Optional.* CallbackQuery handler.
        :param submenu: *Optional.* Submenu name.
        :param items_per_row: *Optional.* Number of items on each row; 1-8.

        .. note::

            Use ``submenu`` to open submenu with button payload; use ``action``
            to execute callback with button payload.
        """
        if ((action is None and submenu is None)
                or (action is not None and submenu is not None)):
            raise ValueError("you must use `submenu` OR `action`")

        self._items_per_row = max(min(items_per_row, 8), 1)
        self._items: ChoiceItems | ChoiceItemsFn = (  # yapf: disable
            items
            if callable(items)
            else [
                (str(i), str(i)) if isinstance(i, str | int) else i
                for i in items
            ]
        )
        self._submenu = submenu
        self._action = action
        self._payload: str | None = None

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Choice) and (  # yapf: disable
            other is self or (
                other._action is self._action
                and other._items == self._items
            )
        )

    def _init(
        self,
        menu: Menu,
        router: OnCallbackQuery,
        payload: Payload,
    ) -> None:
        if self._payload is not None:
            raise ValueError(f"{self!r} must be unique.")

        if self._submenu is not None:
            self._payload = f"{menu.get_submenu(self._submenu)._prefix}##"
            return

        self._payload = payload.get()
        handler = Handler(
            cast("HandlerCallable[CallbackQuery]", self._action),
            update_type=router._update_type,
            filters=[CallbackPayload(self._payload)],
        )

        if handler in router._handlers:
            raise ValueError(f"{self!r} `action` must be unique.")

        router.add_handler(handler)

    async def _build(self, q: Query) -> list[list[InlineKeyboardButton]]:
        return build_rows(
            await it(q) if callable(it := self._items) else it,
            self._payload,  # type: ignore[arg-type]
            self._items_per_row,
        )


class Paging(AbstractRowBuilder):
    """Paging row builder."""

    def __init__(
        self,
        items: PagingItemsFn,
        action: HandlerCallable[CallbackQuery] | None = None,
        submenu: str | None = None,
        items_per_page: int = 5,
        items_per_row: int = 1,
        prev_btn_title: str = "Prev",
        next_btn_title: str = "Next",
    ) -> None:
        """
        :param items: A callable that returns the list of items for current page.
        :param action: *Optional.* CallbackQuery handler.
        :param submenu: *Optional.* Submenu name.
        :param items_per_page: *Optional.* Number of items per page; 1-100.
        :param items_per_row: *Optional.* Number of items per row; 1-8.
        :param prev_btn_title: *Optional.* Title for button 'Previous page'.
        :param next_btn_title: *Optional.* Title for button 'Next page'.

        .. note::

            Use ``submenu`` to open submenu with button payload; use ``action``
            to execute callback with button payload.
        """  # noqa: E501
        self._items = items
        self._items_per_page = max(min(items_per_page, 100), 1)
        self._items_per_row = max(min(items_per_row, 8), 1)
        self._payload: str | None = None
        self._nav_payload: str | None = None
        self._prev_btn_title = prev_btn_title
        self._next_btn_title = next_btn_title
        if ((action is None and submenu is None)
                or (action is not None and submenu is not None)):
            raise ValueError("you must use `submenu` OR `action`")
        self._submenu = submenu
        self._action = action

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Paging) and (  # yapf: disable
            other is self or (
                other._action is self._action
                and other._items is self._items
                and other._items_per_row == self._items_per_row
                and other._items_per_page == self._items_per_page
            )
        )

    def _init(
        self,
        menu: Menu,
        router: OnCallbackQuery,
        payload: Payload,
    ) -> None:
        if self._payload is not None:
            raise ValueError(f"{self!r} must be unique.")

        if self._submenu is not None:
            self._payload = f"{menu.get_submenu(self._submenu)._prefix}##"

        else:
            self._payload = payload.get()
            handler = Handler(
                cast("HandlerCallable[CallbackQuery]", self._action),
                filters=[CallbackPayload(self._payload)],
                update_type=router._update_type,
            )
            if handler in router._handlers:
                raise ValueError(f"{self!r} `action` must be unique.")
            router.add_handler(handler)

        self._nav_payload = f"{menu._prefix}##"

    @staticmethod
    def _parse_offset(payload: str) -> int:
        offset = 0
        if offset_list := payload.rsplit("#", maxsplit=1)[1:]:
            try:
                offset = int(offset_list[0])
            except ValueError:
                pass
        return max(0, offset)

    async def _build(self, q: Query) -> list[list[InlineKeyboardButton]]:
        offset = self._parse_offset(getattr(q, "data", ""))
        items, is_last = await self._items(q, offset, self._items_per_page)
        markup = build_rows(
            items,
            self._payload,  # type: ignore[arg-type]
            self._items_per_row,
        )
        nav = []
        # add button `Previous page`
        if offset > 0:
            extra = offset - self._items_per_page
            nav.append(
                InlineKeyboardButton(
                    text=self._prev_btn_title,
                    callback_data=f"{self._nav_payload}#{extra}",
                )
            )
        # add button `Next page`
        if not is_last:
            extra = offset + self._items_per_page
            nav.append(
                InlineKeyboardButton(
                    text=self._next_btn_title,
                    callback_data=f"{self._nav_payload}#{extra}",
                )
            )
        if nav:
            markup.append(nav)
        return markup



import pytest

from yatbaf_menu.button import URL
from yatbaf_menu.button import Back
from yatbaf_menu.menu import _parse_markup
from yatbaf_menu.row import Choice
from yatbaf_menu.row import Row


@pytest.mark.parametrize(
    "markup,expected",
    [
        (
            [
                Choice(action=(a := object()), items=[1, 2, 3]),
                Back(title="back"),
            ],
            [
                Choice(action=a, items=[1, 2, 3]),
                Row([Back(title="back")]),
            ],
        ),
        (
            [
                [
                    URL(url="example.com", title="example 1"),
                    URL(url="example.com", title="example 2"),
                ],
                [
                    URL(url="example.com", title="example 3"),
                    URL(url="example.com", title="example 4"),
                ],
            ],
            [
                Row([
                    URL(url="example.com", title="example 1"),
                    URL(url="example.com", title="example 2"),
                ]),
                Row([
                    URL(url="example.com", title="example 3"),
                    URL(url="example.com", title="example 4"),
                ]),
            ],
        ),
        (
            [
                URL(url="example.com", title="example 1"),
                URL(url="example.com", title="example 2"),
                URL(url="example.com", title="example 3"),
                URL(url="example.com", title="example 4"),
            ],
            [
                Row([URL(url="example.com", title="example 1")]),
                Row([URL(url="example.com", title="example 2")]),
                Row([URL(url="example.com", title="example 3")]),
                Row([URL(url="example.com", title="example 4")]),
            ],
        ),
        (
            [
                [
                    Row([
                        URL(url="example.com", title="example 1"),
                        URL(url="example.com", title="example 2"),
                    ]),
                ],
                URL(url="example.com", title="example 3"),
                URL(url="example.com", title="example 4"),
            ],
            [
                Row([
                    URL(url="example.com", title="example 1"),
                    URL(url="example.com", title="example 2"),
                ]),
                Row([URL(url="example.com", title="example 3")]),
                Row([URL(url="example.com", title="example 4")]),
            ],
        ),
        (
            [[Choice(submenu="qwe", items=[1, 2, 3])]],
            [Choice(submenu="qwe", items=[1, 2, 3])],
        ),
    ]
)
def test_parse_markup(markup, expected):
    assert _parse_markup(markup) == expected


def test_error():
    with pytest.raises(ValueError):
        _parse_markup([
            [
                Choice(submenu="qwe", items=[1, 2, 3]),
                Choice(submenu="ewq", items=[3, 2, 1]),
            ],
        ])



import pytest

from yatbaf.types import InlineKeyboardButton
from yatbaf_menu.button import URL


@pytest.mark.asyncio
async def test_build(callback_query):
    button = URL(title="link", url="url")
    result = await button._build(callback_query)
    assert result == InlineKeyboardButton(text="link", url="url")


@pytest.mark.asyncio
async def test_build_dynamic_url(callback_query):

    async def _url(_):
        return "dyn_url"

    button = URL(title="link", url=_url)
    result = await button._build(callback_query)
    assert result == InlineKeyboardButton(text="link", url="dyn_url")



import pytest

from yatbaf import OnCallbackQuery
from yatbaf.exceptions import SkipRouterException
from yatbaf_menu.button import URL
from yatbaf_menu.menu import Menu
from yatbaf_menu.middleware import CutPayloadMiddileware
from yatbaf_menu.middleware import InjectMenuMiddleware
from yatbaf_menu.utils import _build_base_router
from yatbaf_menu.utils import _build_fallback_router
from yatbaf_menu.utils import _build_menu_router
from yatbaf_menu.utils import _build_root_router


@pytest.fixture
def menu(prefix):
    result = Menu(
        title="menu title",
        name="test_menu",
        buttons=[URL(title="button", url="example.com")],
    )
    result._prefix = prefix
    return result


@pytest.mark.asyncio
async def test_build_menu_router(menu):
    router = _build_menu_router(menu)

    assert router._middleware[0].obj is InjectMenuMiddleware
    assert router._middleware[0].is_local
    assert router._middleware[0].is_handler
    assert router._handlers[0]._filters[0].payload == f"{menu._prefix}##"


@pytest.mark.asyncio
async def test_build_base_router(menu, callback_query):
    version = "00"
    router = _build_base_router(menu, version)

    assert not router._handlers
    assert router._middleware[0].obj is CutPayloadMiddileware
    assert router._middleware[0].is_handler
    assert router._guards
    guard = router._guards[0]

    callback_query.data = f"{menu._prefix[:2]}{version}ee"
    guard(callback_query)

    callback_query.data = f"{menu._prefix[:2]}01ee"
    with pytest.raises(SkipRouterException) as err:
        guard(callback_query)
    assert err.value.skip_nested


def test_build_fallback_router(menu):
    router = _build_fallback_router(menu, "updated")

    assert not router._guards
    assert router._middleware
    assert router._middleware[0].obj is InjectMenuMiddleware
    assert router._middleware[0].is_handler
    assert router._middleware[0].is_local
    assert router._middleware[0].params["menu"] is menu


@pytest.mark.asyncio
async def test_build_root_router(menu, callback_query):
    base = OnCallbackQuery(handlers=[object()])
    fallback = OnCallbackQuery(handlers=[object()])

    router = _build_root_router(
        menu=menu,
        base_router=base,
        fallback_router=fallback,
    )

    assert router._routers
    assert router._routers[0] is base
    assert router._routers[1] is fallback

    assert not router._middleware
    assert router._guards
    guard = router._guards[0]

    callback_query.data = f"{menu._prefix}qwer123"
    guard(callback_query)

    callback_query.data = "any"
    with pytest.raises(SkipRouterException) as err:
        guard(callback_query)
    assert err.value.skip_nested

    callback_query.data = None
    with pytest.raises(SkipRouterException) as err:
        guard(callback_query)
    assert err.value.skip_nested



import pytest

from yatbaf.types import InlineKeyboardButton
from yatbaf_menu import Menu
from yatbaf_menu import Submenu


def test_init_no_submenu(router, payload):
    menu = Menu(
        title="menu",
        name="menu",
        buttons=[
            button := Submenu(title="submenu", menu="submenu"),
        ],
    )
    with pytest.raises(ValueError):
        button._init(menu, router, payload)


def test_init(router, prefix, payload):
    menu = Menu(
        title="menu",
        name="menu",
        buttons=[
            button := Submenu(title="submenu", menu="submenu"),
        ],
        submenu=[
            submenu := Menu(
                title="submenu",
                name="submenu",
                back_btn_title="back",
            ),
        ],
    )
    submenu._prefix = prefix
    button._init(menu, router, payload)
    assert button._payload == f"{submenu._prefix}##"


def test_init_duplicate(router, prefix, payload):
    menu = Menu(
        title="menu",
        name="menu",
        buttons=[
            button := Submenu(title="submenu", menu="submenu"),
        ],
        submenu=[
            Menu(
                title="submenu",
                name="submenu",
                back_btn_title="back",
            ),
        ]
    )
    button._payload = f"{prefix}##"
    with pytest.raises(ValueError):
        button._init(menu, router, payload)


@pytest.mark.asyncio
async def test_build(callback_query, prefix):
    button = Submenu(title="submenu", menu="submenu")
    button._payload = f"{prefix}##"
    result = await button._build(callback_query)
    assert result == InlineKeyboardButton(
        text="submenu",
        callback_data=button._payload,
    )


@pytest.mark.asyncio
async def test_build_not_visible(callback_query, prefix):

    async def _visible(_):
        return False

    button = Submenu(title="submenu", menu="submenu", show=_visible)
    button._payload = f"{prefix}##"
    result = await button._build(callback_query)
    assert result is None



from __future__ import annotations

__all__ = ("Payload",)

from itertools import product
from string import ascii_letters
from string import digits


class Payload:
    __slots__ = (
        "_bucket",
        "_prefix",
    )

    def __init__(self, prefix: str = "") -> None:
        self._bucket = product(f"{ascii_letters}{digits}", repeat=2).__next__
        self._prefix = prefix

    @property
    def prefix(self) -> str:
        return self._prefix

    def get(self) -> str:
        return f"{self._prefix}{''.join(self._bucket())}"



import pytest

from yatbaf.types import InlineKeyboardButton
from yatbaf.types import InlineKeyboardMarkup
from yatbaf_menu.button import URL
from yatbaf_menu.menu import Menu


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "buttons,markup",
    [
        [
            [
                URL(url="example.com", title="example 1"),
                URL(url="example.com", title="example 2"),
                URL(url="example.com", title="example 3"),
                URL(url="example.com", title="example 4"),
            ],
            [
                [InlineKeyboardButton(url="example.com", text="example 1")],
                [InlineKeyboardButton(url="example.com", text="example 2")],
                [InlineKeyboardButton(url="example.com", text="example 3")],
                [InlineKeyboardButton(url="example.com", text="example 4")],
            ],
        ],
        [
            [
                [
                    URL(url="example.com", title="example 1"),
                    URL(url="example.com", title="example 2"),
                    URL(url="example.com", title="example 3"),
                ],
                URL(url="example.com", title="example 4"),
            ],
            [
                [
                    InlineKeyboardButton(url="example.com", text="example 1"),
                    InlineKeyboardButton(url="example.com", text="example 2"),
                    InlineKeyboardButton(url="example.com", text="example 3"),
                ],
                [InlineKeyboardButton(url="example.com", text="example 4")],
            ],
        ],
        [
            [
                [
                    URL(url="example.com", title="example 1"),
                    URL(url="example.com", title="example 2"),
                ],
                [
                    URL(url="example.com", title="example 3"),
                    URL(url="example.com", title="example 4"),
                ],
            ],
            [
                [
                    InlineKeyboardButton(url="example.com", text="example 1"),
                    InlineKeyboardButton(url="example.com", text="example 2")
                ],
                [
                    InlineKeyboardButton(url="example.com", text="example 3"),
                    InlineKeyboardButton(url="example.com", text="example 4")
                ],
            ],
        ],
    ]
)
async def test_build_markup(buttons, markup):
    menu = Menu(title="test", name="test", buttons=buttons)
    assert await menu._get_markup(None) == InlineKeyboardMarkup(markup)



from __future__ import annotations

__all__ = ("MenuNav",)

from typing import TYPE_CHECKING
from typing import cast

from yatbaf.exceptions import MethodInvokeError

if TYPE_CHECKING:
    from yatbaf.types import CallbackQuery
    from yatbaf.types import Message

    from .menu import Menu


class MenuNav:
    """Menu navigation.

    Usage::

        async def button_action(q: CallbackQuery) -> None:
            nav = q.ctx["menu"]
            ...

    .. important::

        Make sure the ``q.message`` is :class:`~yatbaf.types.Message` before
        use these methods.

    """

    __slots__ = (
        "_query",
        "_menu",
    )

    def __init__(self, menu: Menu, query: CallbackQuery, /) -> None:
        """
        :param menu: Current menu.
        :param query: Current query.
        """
        self._query = query
        self._menu = menu

    @property
    def menu(self) -> Menu:
        """Current :class:`~yatbaf_menu.menu.Menu` instance."""
        return self._menu

    async def _open_menu(self, menu: Menu) -> None:
        message = cast("Message", self._query.message)
        params = await menu.get_message_params(self._query)
        await message.edit(**params)

    async def goto(self, path: str) -> None:
        """Use this method to go to menu by path.

        :param path: Path to menu relative to main menu. Separated by dots.
        """
        menu = self._menu.get_menu(path)
        await self._open_menu(menu)

    async def open_previous(self) -> None:
        """Use this method to go to previous menu."""
        menu = self._menu._parent
        if menu is None:
            return
        await self._open_menu(menu)

    async def open_main(self) -> None:
        """Use this method to go to main menu."""
        menu = self._menu.get_root()
        await self._open_menu(menu)

    async def open_submenu(self, name: str) -> None:
        """Use this method to go to submenu.

        :param name: Submenu name.
        """
        menu = self._menu.get_submenu(name)
        await self._open_menu(menu)

    async def open(self) -> None:
        await self._open_menu(self._menu)

    async def refresh(self) -> None:
        """Use this method to refresh current menu."""
        await self._open_menu(self._menu)

    async def close(self) -> None:
        """Use this method to close the menu."""
        message = cast("Message", self._query.message)
        try:
            await message.delete()
        except MethodInvokeError:
            await message.edit_reply_markup()



import pytest

from yatbaf.types import InlineKeyboardButton
from yatbaf_menu.menu import Menu
from yatbaf_menu.row import Choice

PAYLOAD = "~a00a"


def test_no_action_and_submenu():
    with pytest.raises(ValueError):
        Choice(items=[1, 2, 3])


def test_action_and_submenu():
    with pytest.raises(ValueError):
        Choice(
            action=object(),
            submenu="menu",
            items=[1, 2, 3],
        )


def test_init_submenu(prefix, router, payload):
    menu = Menu(
        title="menu",
        name="menu",
        submenu=[
            submenu := Menu(
                title="submenu",
                name="submenu",
                back_btn_title="back",
            ),
        ],
        buttons=[
            choice := Choice(
                submenu="submenu",
                items=[1, 2, 3],
            ),
        ],
    )
    menu._prefix = prefix
    submenu._prefix = f"{prefix[:-2]}bb"
    choice._init(menu, router, payload)
    assert choice._payload == f"{submenu._prefix}##"
    assert not router._handlers


def test_init_action(prefix, router, payload):
    action = object()
    menu = Menu(
        title="menu",
        name="menu",
        buttons=[
            choice := Choice(
                action=action,
                items=[("1", "1"), ("2", "2")],
                items_per_row=1,
            )
        ],
    )
    menu._prefix = prefix
    choice._init(menu, router, payload)
    assert choice._payload is not None
    assert router._handlers[0]._fn is action
    assert router._handlers[0]._filters[0].payload == choice._payload


def test_init_duplicate(prefix, router, payload):
    action = object()
    menu = Menu(
        title="menu",
        name="menu",
        buttons=[
            choice := Choice(
                action=action,
                items=[("1", "1"), ("2", "2")],
                items_per_row=1,
            ),
        ]
    )
    menu._prefix = prefix
    choice._payload = "qwe"
    with pytest.raises(ValueError):
        choice._init(menu, router, payload)


def test_init_duplicate_action(prefix, router, payload):
    action = object()
    menu = Menu(
        title="menu",
        name="menu",
        buttons=[
            choice := Choice(
                action=action,
                items=[("1", "1"), ("2", "2")],
            )
        ]
    )
    menu._prefix = prefix
    router.add_handler(action)
    with pytest.raises(ValueError):
        choice._init(menu, router, payload)


@pytest.mark.asyncio
async def test_build():

    async def _items(_):
        return [("1", "1"), ("2", "2"), ("3", "3")]

    choice = Choice(action=object(), items=_items, items_per_row=1)
    choice._payload = PAYLOAD
    assert choice._items is _items

    result = await choice._build(None)
    assert result == [
        [InlineKeyboardButton(text="1", callback_data=f"{PAYLOAD}1")],
        [InlineKeyboardButton(text="2", callback_data=f"{PAYLOAD}2")],
        [InlineKeyboardButton(text="3", callback_data=f"{PAYLOAD}3")],
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "ipr,expected",
    [
        (
            1,
            [
                [InlineKeyboardButton(text="1", callback_data=f"{PAYLOAD}1")],
                [InlineKeyboardButton(text="2", callback_data=f"{PAYLOAD}2")],
                [InlineKeyboardButton(text="3", callback_data=f"{PAYLOAD}3")],
            ]
        ),
        (
            3,
            [
                [
                    InlineKeyboardButton(text="1", callback_data=f"{PAYLOAD}1"),
                    InlineKeyboardButton(text="2", callback_data=f"{PAYLOAD}2"),
                    InlineKeyboardButton(text="3", callback_data=f"{PAYLOAD}3")
                ],
            ]
        ),
    ]
)
async def test_items_per_row(ipr, expected):
    choice = Choice(action=object(), items=[1, 2, 3], items_per_row=ipr)
    choice._payload = PAYLOAD
    assert choice._items == [("1", "1"), ("2", "2"), ("3", "3")]

    result = await choice._build(None)
    assert result == expected



from __future__ import annotations

from typing import TYPE_CHECKING

from .nav import MenuNav

if TYPE_CHECKING:
    from yatbaf.types import CallbackQuery
    from yatbaf.typing import HandlerCallable

    from .menu import Menu


class CutPayloadMiddileware:
    __slots__ = ("_fn",)

    def __init__(self, fn: HandlerCallable[CallbackQuery]) -> None:
        self._fn = fn

    async def __call__(self, q: CallbackQuery) -> None:
        # root menu (2ch) + version (2ch) + menu (2ch) + button (2ch)
        q.data = q.data[8:]  # type: ignore[index]
        await self._fn(q)


class InjectMenuMiddleware:
    __slots__ = (
        "_fn",
        "_menu",
    )

    def __init__(self, fn: HandlerCallable[CallbackQuery], menu: Menu) -> None:
        self._fn = fn
        self._menu = menu

    async def __call__(self, q: CallbackQuery) -> None:
        q.ctx["menu"] = MenuNav(self._menu, q)
        await self._fn(q)
